import os
import unittest
import asyncio

from lionagi.utils.api_util import BaseAPIService, StatusTracker, AsyncQueue, RateLimiter

class MockAPIService(BaseAPIService):
    async def call_api_endpoint(self):
        pass  # Minimal implementation for testing

    # Optionally, override other methods if specific behavior needs to be tested

# class TestBaseAPIService(unittest.TestCase):
#
    # def setUp(self):
    #     self.api_service = MockAPIService(
    #         api_key="test_api_key",
    #         token_encoding_name="test_encoding",
    #         max_attempts=3,
    #         status_tracker=StatusTracker(),
    #         rate_limiter=None,
    #         queue=AsyncQueue()
    #     )
    #     self.test_filename = "test_log.jsonl"
    #
    # def tearDown(self):
    #     if os.path.exists(self.test_filename):
    #         os.remove(self.test_filename)
    #
    # def test_handle_error(self):
    #     error = Exception("Test error")
    #     payload = {"data": "test"}
    #     metadata = {"attempt": 1}
    #     self.api_service.handle_error(error, payload, metadata, self.test_filename)
    #
    #     self.assertEqual(self.api_service.status_tracker.num_tasks_failed, 1)
    #     self.assertEqual(self.api_service.status_tracker.num_tasks_in_progress, -1)
    #
    #     # Check if error is logged correctly
    #     with open(self.test_filename, "r") as file:
    #         logged_data = file.readline()
    #         self.assertIn("Test error", logged_data)
    #         self.assertIn("test", logged_data)
    #
    # def test_api_endpoint_from_url(self):
    #     test_url = "https://example.com/v1/test_endpoint"
    #     endpoint = self.api_service.api_endpoint_from_url(test_url)
    #     self.assertEqual(endpoint, "test_endpoint")
    #
    # def test_task_id_generator_function(self):
    #     id_gen = self.api_service.task_id_generator_function()
    #     self.assertEqual(next(id_gen), 0)
    #     self.assertEqual(next(id_gen), 1)


class TestAsyncQueue(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.queue = AsyncQueue()
        self.results = []

    async def asyncTearDown(self):
        await self.queue.stop()
        # Ensure queue processing has been stopped
        await asyncio.sleep(0.1)

    async def test_queue_processing(self):
        async def dummy_task(item):
            await asyncio.sleep(0.1)
            self.results.append(item)

        processing_task = asyncio.create_task(
            self.queue.process_requests(dummy_task)
        )

        for item in range(5):
            await self.queue.enqueue(item)

        # Signal the end of processing by enqueuing None
        await self.queue.enqueue(None)

        # Wait until the processing task has stopped
        await processing_task

        # Check if all items were processed
        self.assertEqual(self.results, [0, 1, 2, 3, 4])



class MockRateLimiter(RateLimiter):
    def __init__(self, max_requests_per_minute, max_tokens_per_minute):
        super().__init__(max_requests_per_minute, max_tokens_per_minute)

    async def rate_limit_replenisher(self):
        # Replenish logic (simplified for testing)
        await asyncio.sleep(0.1)  # Fast replenish for testing
        self.available_request_capacity = self.max_requests_per_minute
        self.available_token_capacity = self.max_tokens_per_minute

    def calculate_num_token(self, request_details):
        # Simplified token calculation for testing
        return len(request_details)



class TestMockRateLimiter(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.rate_limiter = MockRateLimiter(max_requests_per_minute=100, max_tokens_per_minute=200)

    async def test_rate_limit_replenishment(self):
        self.rate_limiter.available_request_capacity = 0
        self.rate_limiter.available_token_capacity = 0

        await asyncio.sleep(0.2)  # Wait for replenishment to occur

        self.assertEqual(self.rate_limiter.available_request_capacity, 100)
        self.assertEqual(self.rate_limiter.available_token_capacity, 200)

    def test_token_calculation(self):
        request_details = "test_request"
        tokens = self.rate_limiter.calculate_num_token(request_details)

        self.assertEqual(tokens, len(request_details))

class TestStatusTracker(unittest.TestCase):

    def setUp(self):
        self.tracker = StatusTracker()

    def test_task_started(self):
        self.tracker.num_tasks_started += 1
        self.assertEqual(self.tracker.num_tasks_started, 1)

    def test_task_in_progress(self):
        self.tracker.num_tasks_in_progress += 1
        self.assertEqual(self.tracker.num_tasks_in_progress, 1)

    def test_task_succeeded(self):
        self.tracker.num_tasks_succeeded += 1
        self.assertEqual(self.tracker.num_tasks_succeeded, 1)

    def test_task_failed(self):
        self.tracker.num_tasks_failed += 1
        self.assertEqual(self.tracker.num_tasks_failed, 1)

    def test_rate_limit_errors(self):
        self.tracker.num_rate_limit_errors += 1
        self.assertEqual(self.tracker.num_rate_limit_errors, 1)

    def test_api_errors(self):
        self.tracker.num_api_errors += 1
        self.assertEqual(self.tracker.num_api_errors, 1)

    def test_other_errors(self):
        self.tracker.num_other_errors += 1
        self.assertEqual(self.tracker.num_other_errors, 1)

if __name__ == '__main__':
    unittest.main()



"""   
Copyright 2023 HaiyangLi <ocean@lionagi.ai>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import asyncio
import copy
import csv
import json
import os
import re
import tempfile
import time
import hashlib
from datetime import datetime
from collections.abc import Generator, Iterable, MutableMapping
from typing import Any, Callable, Dict, List, Optional, Tuple, Union


def _flatten_dict(input: Dict[str, Any], parent_key: str = '', 
                  sep: str = '_') -> Generator[tuple[str, Any], None, None]:
    """
    Recursively flattens a nested dictionary into a flat dictionary.

    This generator function traverses a nested dictionary, concatenating nested keys 
    into a single key using the provided separator. It handles both nested dictionaries
    and lists within the dictionary.

    Parameters:
        input (Dict[str, Any]): The dictionary to be flattened.
        parent_key (str, optional): Initial key for nested structures. Defaults to ''.
        sep (str, optional): Separator for concatenated key. Defaults to '_'.

    Yields:
        Generator[tuple[str, Any], None, None]: Tuples of flattened keys and their values.

    Examples:
        >>> example_dict = {"a": 1, "b": {"c": 2, "d": {"e": 3}}}
        >>> list(_flatten_dict(example_dict))
        [('a', 1), ('b_c', 2), ('b_d_e', 3)]

        >>> example_dict_with_list = {"a": 1, "b": [{"c": 2}, {"d": 3}]}
        >>> list(_flatten_dict(example_dict_with_list))
        [('a', 1), ('b_0_c', 2), ('b_1_d', 3)]
    """
    for key, value in input.items():
        composed_key = f'{parent_key}{sep}{key}' if parent_key else key
        if isinstance(value, MutableMapping):
            yield from _flatten_dict(value, composed_key, sep=sep)
        elif isinstance(value, list):
            for i, item in enumerate(value):
                if isinstance(item, MutableMapping):
                    yield from _flatten_dict(item, f'{composed_key}{sep}{i}', sep=sep)
                else:
                    yield (f'{composed_key}{sep}{i}', item)
        else:
            yield (composed_key, value)

def to_flat_dict(input: Dict[str, Any], parent_key: str = '', sep: str = '_') -> Dict[str, Any]:
    """
    Transforms a nested dictionary into a flat dictionary.

    This function flattens a dictionary with potentially nested structures, concatenating
    nested keys using the specified separator. It leverages the '_flatten_dict' generator
    function for efficient processing.

    Parameters:
        input (Dict[str, Any]): The nested dictionary to flatten.
        parent_key (str, optional): Initial key for nested structures. Defaults to ''.
        sep (str, optional): Separator for concatenated keys. Defaults to '_'.

    Returns:
        Dict[str, Any]: The resulting flat dictionary.

    Example:
        >>> example_nested_dict = {'a': 1, 'b': {'c': 2}}
        >>> to_flat_dict(example_nested_dict)
        {'a': 1, 'b_c': 2}
    """
    return dict(_flatten_dict(input, parent_key, sep))

def _flatten_list(input: List[Any], dropna: bool = True) -> Generator[Any, None, None]:
    """
    Recursively flattens a nested list into a single-level list.

    This generator function iterates through a list, flattening nested lists into a single 
    level. It provides an option to exclude 'None' values from the output, enhancing its utility.

    Parameters:
        input (List[Any]): The list to be flattened.
        dropna (bool, optional): Flag to indicate whether 'None' values should be omitted. Defaults to True.

    Yields:
        Generator[Any, None, None]: Elements from the flattened list.

    Example:
        >>> example_nested_list = [1, [2, 3, None], 4]
        >>> list(_flatten_list(example_nested_list))
        [1, 2, 3, 4]
    """
    for element in input:
        if isinstance(element, list):
            yield from _flatten_list(element, dropna)
        elif element is not None or not dropna:
            yield element

def to_list(input: Union[Iterable, Any], flatten_dict: bool = False, flat: bool = True, 
            dropna: bool = True, parent_key: str = '', sep: str = '_') -> List[Any]:
    """
    Converts a given input into a list, optionally flattening nested structures.

    This function converts various inputs into a list, with optional handling for 
    flattening nested lists and dictionaries, and dropping 'None' values. If 'flatten_dict'
    is True and the input is a dictionary, it is flattened before conversion.

    Parameters:
        input (Union[Iterable, Any]): The input to convert.
        flatten_dict (bool, optional): Flatten dictionary input. Defaults to False.
        flat (bool, optional): Flatten the output list. Defaults to True.
        dropna (bool, optional): Drop 'None' values during flattening. Defaults to True.

    Raises:
        ValueError: If input is None, callable, or unconvertible to a list.

    Returns:
        List[Any]: The resulting list after applying transformations.

    Example:
        >>> to_list({'a': 1, 'b': [2, 3]}, flatten_dict=True)
        [{'a': 1}, {'b_0': 2}, {'b_1': 3}]
    """

    if input is None:
        raise ValueError("Unsupported type: None are not convertible to a list.")
    
    try:
        if isinstance(input, dict):
            if flatten_dict:
                input = dict(_flatten_dict(input, parent_key, sep))  # Flatten and convert to dictionary first
                return [{k: v} for k, v in input.items()]
            output = [input]
        elif isinstance(input, Iterable) and not isinstance(input, str):
            output = list(input)
        else:
            output = [input]
        
        if flat:  # Flatten if necessary
            output = list(_flatten_list(output, dropna))
        return output
    
    except TypeError as e:
        raise ValueError(f"Unable to convert input to list. Error: {e}")

def str_to_num(input: str, 
               upper_bound: Optional[Union[int, float]] = None, 
               lower_bound: Optional[Union[int, float]] = None, 
               num_type: type = int, 
               precision: Optional[int] = None) -> Union[int, float]:
    """
    Converts found numeric value in a string to a specified type.

    This function searches a string for numeric values and converts it to an integer 
    or a float, based on the specified type. It also validates the converted value 
    against optional upper and lower bounds.

    Parameters:
        input (str): String to search for numeric values.
        upper_bound (Optional[Union[int, float]]): Upper limit for the numeric value. None for no limit.
        lower_bound (Optional[Union[int, float]]): Lower limit for the numeric value. None for no limit.
        num_type (type): Desired type for the numeric value (int or float).
        precision (Optional[int]): Decimal places for rounding if float. None for no rounding.

    Raises:
        ValueError: If no numeric values are found, for conversion errors, or if the number is out of bounds.

    Returns:
        Union[int, float]: The converted numeric value, if within specified bounds.

    Example:
        >>> str_to_num("Temperature is -5.6 degrees", num_type=float, precision=1)
        -5.6
        >>> str_to_num("Value is approximately 200", upper_bound=150)
        ValueError: Number 200 is greater than the upper bound of 150.
    """
    numbers = re.findall(r'-?\d+\.?\d*', input)
    if not numbers:
        raise ValueError(f"No numeric values found in the string: {input}")
    
    try:
        numbers = numbers[0]
        
        if num_type is int: 
            numbers = int(float(numbers))
        elif num_type is float:
            numbers = round(float(numbers), precision) if precision is not None else float(numbers)
        else:
            raise ValueError(f"Invalid number type: {num_type}")
        
        if upper_bound is not None and numbers > upper_bound:
            raise ValueError(f"Number {numbers} is greater than the upper bound of {upper_bound}.")
        if lower_bound is not None and numbers < lower_bound:
            raise ValueError(f"Number {numbers} is less than the lower bound of {lower_bound}.")

        return numbers
    except ValueError as e:
        raise ValueError(f"Error converting string to number: {e}")

def make_copy(input: Any, n: int) -> Any:
    """
    Creates a deep copy of the input object a specified number of times.

    This function makes deep copies of the provided input. If the number of copies ('n') 
    is greater than 1, a list of deep copies is returned. For a single copy, it returns 
    the copy directly.

    Parameters:
        input (Any): The object to be copied.
        n (int): The number of deep copies to create.

    Raises:
        ValueError: If 'n' is not a positive integer.

    Returns:
        Any: A deep copy of 'input' or a list of deep copies if 'n' > 1.

    Example:
        >>> sample_dict = {'key': 'value'}
        >>> make_copy(sample_dict, 2)
        [{'key': 'value'}, {'key': 'value'}]
    """
    if not isinstance(n, int) or n < 1:
        raise ValueError(f"'n' must be a positive integer: {n}")
    return copy.deepcopy(input) if n == 1 else [copy.deepcopy(input) for _ in range(n)]

def to_temp(input: Any, 
            flatten_dict: bool = False, 
            flat: bool = False, 
            dropna: bool = False):
    """
    Converts input to a list and writes it to a temporary file in JSON format, with flattening options.

    This function serializes data to a temporary JSON file, useful for transient storage or testing. 
    It includes options to flatten the input if it contains dictionaries or lists.

    Parameters:
        input (Any): The data to be converted and written to a file.
        flatten_dict (bool, optional): Flatten dictionaries in the input. Defaults to False.
        flat (bool, optional): Flatten lists in the input. Defaults to False.
        dropna (bool, optional): Exclude 'None' values during flattening. Defaults to False.

    Raises:
        TypeError: If the input is not JSON serializable.

    Example:
        >>> temp_file = to_temp({'a': 1, 'b': [2, 3]}, flatten_dict=True)
        >>> temp_file.name  # Doctest: +ELLIPSIS
        '/var/folders/.../tmp...'
    """
    input = to_list(input, flatten_dict, flat, dropna)
    
    temp_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
    try:
        json.dump(input, temp_file)
    except TypeError as e:
        temp_file.close()  # Ensuring file closure before raising error
        raise TypeError(f"Data provided is not JSON serializable: {e}")
    temp_file.close()
    return temp_file

def to_csv(input: List[Dict[str, Any]],
           filepath: str,
           file_exist_ok: bool = False) -> None:
    """
    Writes a list of dictionaries to a CSV file, with dictionary keys as headers.

    This function writes a list of dictionaries to a CSV file. It checks if the file exists 
    and handles file creation based on the 'file_exist_ok' flag.

    Parameters:
        input (List[Dict[str, Any]]): Data to write to the CSV file.
        filepath (str): Path of the output CSV file.
        file_exist_ok (bool, optional): Create the file if it doesn't exist. Defaults to False.

    Raises:
        FileExistsError: If the file already exists and 'file_exist_ok' is False.

    Example:
        >>> data = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
        >>> to_csv(data, 'people.csv')
    """

    if not os.path.exists(os.path.dirname(filepath)) and os.path.dirname(filepath) != '':
        if file_exist_ok:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
        else:
            raise FileNotFoundError(f"The directory {os.path.dirname(filepath)} does not exist.")

    with open(filepath, 'w', newline='') as csv_file:
        writer = csv.DictWriter(csv_file, fieldnames=input[0].keys())
        writer.writeheader()
        writer.writerows(input)



def append_to_jsonl(data: Any, filepath: str) -> None:
    """
    Appends data to a JSON lines (jsonl) file.

    Serializes given data to a JSON-formatted string and appends it to a jsonl file. 
    Useful for logging or data collection where entries are added incrementally.

    Parameters:
        data (Any): Data to be serialized and appended.
        filepath (str): Path to the jsonl file.

    Example:
        >>> append_to_jsonl({"key": "value"}, "data.jsonl")
        # Appends {"key": "value"} to 'data.jsonl'
    """
    json_string = json.dumps(data)
    with open(filepath, "a") as f:
        f.write(json_string + "\n")

def hold_call(input: Any, 
              func: Callable, 
              sleep: int = 0.1, 
              message: Optional[str] = None, 
              ignore_error: bool = False, 
              **kwargs) -> Any:
    """
    Executes a function after a specified delay, handling exceptions optionally.

    Waits for 'sleep' seconds before calling 'func' with 'input'. Handles exceptions by
    printing a message and optionally re-raising them.

    Parameters:
        input (Any): Input to the function.
        func (Callable): Function to execute.
        sleep (int, optional): Time in seconds to wait before calling the function. Defaults to 0.1.
        message (Optional[str], optional): Message to print on exception. Defaults to None.
        ignore_error (bool, optional): If True, ignores exceptions. Defaults to False.
        **kwargs: Additional keyword arguments for the function.

    Returns:
        Any: Result of the function call.

    Raises:
        Exception: Re-raises the exception unless 'ignore_error' is True.

    Example:
        >>> def add_one(x):
        ...     return x + 1
        >>> hold_call(5, add_one, sleep=2)
        6
    """
    try:
        time.sleep(sleep)
        return func(input, **kwargs)
    except Exception as e:
        if message:
            print(f"{message} Error: {e}")
        else:
            print(f"An error occurred: {e}")
        if not ignore_error:
            raise

async def ahold_call(input: Any, 
                     func: Callable, 
                     sleep: int = 5, 
                     message: Optional[str] = None, 
                     ignore_error: bool = False, 
                     **kwargs) -> Any:
    """
    Asynchronously executes a function after a specified delay, handling exceptions optionally.

    Waits for 'sleep' seconds before calling 'func' with 'input'. Handles exceptions by
    printing a message and optionally re-raising them.

    Parameters:
        input (Any): Input to the function.
        func (Callable): Asynchronous function to execute.
        sleep (int, optional): Time in seconds to wait before calling the function. Defaults to 5.
        message (Optional[str], optional): Message to print on exception. Defaults to None.
        ignore_error (bool, optional): If True, ignores exceptions. Defaults to False.
        **kwargs: Additional keyword arguments for the function.

    Returns:
        Any: Result of the asynchronous function call.

    Raises:
        Exception: Re-raises the exception unless 'ignore_error' is True.

    Example:
        >>> async def async_add_one(x):
        ...     return x + 1
        >>> asyncio.run(ahold_call(5, async_add_one, sleep=2))
        6
    """
    try:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"The function {func} must be an asynchronous function.")
        await asyncio.sleep(sleep)
        return await func(input, **kwargs)
    except Exception as e:
        if message:
            print(f"{message} Error: {e}")
        else:
            print(f"An error occurred: {e}")
        if not ignore_error:
            raise

def l_call(input: Any, 
           func: Callable, 
           flatten_dict: bool = False, 
           flat: bool = False, 
           dropna: bool = True) -> List[Any]:
    """
    Applies a function to each element of `input`, after converting it to a list.

    This function converts the `input` to a list, with options to flatten structures 
    and lists, and then applies a given `func` to each element of the list.

    Parameters:
        input (Any): The input to be converted to a list and processed.
        func (Callable): The function to apply to each element of the list.
        flatten_dict (bool, optional): If True, flattens dictionaries in the input. Defaults to False.
        flat (bool, optional): If True, flattens nested lists in the input. Defaults to False.
        dropna (bool, optional): If True, drops None values during flattening. Defaults to True.

    Returns:
        List[Any]: A list containing the results of applying the `func` to each element.

    Raises:
        ValueError: If the `func` cannot be applied to the `input`.

    Example:
        >>> def square(x):
        ...     return x * x
        >>> l_call([1, 2, 3], square)
        [1, 4, 9]
    """
    try:
        lst = to_list(input, flatten_dict, flat, dropna)
        return [func(i) for i in lst]
    except Exception as e:
        raise ValueError(f"Given function cannot be applied to the input. Error: {e}")

async def al_call(input: Any, 
                  func: Callable, 
                  flatten_dict: bool = False, 
                  flat: bool = False, 
                  dropna: bool = True) -> List[Any]:
    """
    Asynchronously applies a function to each element of `input`, after converting it to a list.

    This function converts the `input` to a list, with options to flatten 
    dictionaries and lists, and then applies a given asynchronous `func` to 
    each element of the list asynchronously.

    Parameters:
        input (Any): The input to be converted to a list and processed.
        func (Callable): The asynchronous function to apply to each element of the list.
        flatten_dict (bool, optional): If True, flattens dictionaries in the input. Defaults to False.
        flat (bool, optional): If True, flattens nested lists in the input. Defaults to False.
        dropna (bool, optional): If True, drops None values during flattening. Defaults to True.

    Returns:
        List[Any]: A list containing the results of applying the `func` to each element.

    Raises:
        ValueError: If the `func` cannot be applied to the `input`.

    Example:
        >>> async def async_square(x):
        ...     return x * x
        >>> asyncio.run(al_call([1, 2, 3], async_square))
        [1, 4, 9]
    """
    try:
        lst = to_list(input, flatten_dict, flat, dropna)
        tasks = [func(i) for i in lst]
        return await asyncio.gather(*tasks)
    except Exception as e:
        raise ValueError(f"Given function cannot be applied to the input. Error: {e}")

def m_call(input: Union[Any, List[Any]], 
           func: Union[Callable, List[Callable]], 
           flatten_dict: bool = False, 
           flat: bool = True, 
           dropna: bool = True) -> List[Any]:
    """
    Maps multiple functions to corresponding elements of the input.

    This function applies a list of functions to a list of inputs, with each function
    being applied to its corresponding input element. It asserts that the number of inputs
    and functions are the same and raises an error if they are not.

    Parameters:
        input (Union[Any, List[Any]]): The input or list of inputs to be processed.
        funcs (Union[Callable, List[Callable]]): The function or list of functions to apply.
        flatten_dict (bool, optional): Whether to flatten dictionaries in the input. Defaults to False.
        flat (bool, optional): Whether the output list should be flattened. Defaults to True.
        dropna (bool, optional): Whether to drop None values during flattening. Defaults to True.

    Returns:
        List[Any]: A list containing the results from applying each function to its corresponding input.

    Raises:
        ValueError: If the number of provided inputs and functions are not the same.

    Example:
        >>> def add_one(x):
        ...     return x + 1
        >>> m_call([1, 2], [add_one, add_one])
        [2, 3]
    """
    input = to_list(input, flatten_dict, flat, dropna)
    func = to_list(func)
    assert len(input) == len(func), "The number of inputs and functions must be the same."
    return to_list([l_call(inp, f, flatten_dict, flat, dropna) 
                    for f, inp in zip(func, input)])

async def am_call(input: Union[Any, List[Any]], 
                  func: Union[Callable, List[Callable]], 
                  flatten_dict: bool = False, 
                  flat: bool = True, 
                  dropna: bool = True) -> List[Any]:
    """
    Asynchronously applies multiple functions to corresponding elements of the input.

    This asynchronous function maps a list of functions to a list of inputs, with each 
    function being applied to its corresponding input element asynchronously. It ensures 
    that the number of inputs and functions are the same, raising a `ValueError` if not.

    Parameters:
        input (Union[Any, List[Any]]): The input or list of inputs to be processed.
        funcs (Union[Callable, List[Callable]]): The function or list of functions to apply.
        flatten_dict (bool, optional): Whether to flatten dictionaries in the input. Defaults to False.
        flat (bool, optional): Whether the output list should be flattened. Defaults to True.
        dropna (bool, optional): Whether to drop None values during flattening. Defaults to True.

    Returns:
        List[Any]: A list containing the results from applying each function to its corresponding input.

    Raises:
        ValueError: If the number of inputs and functions do not match.

    Example:
        >>> async def async_add_one(x):
        ...     return x + 1
        >>> asyncio.run(am_call([1, 2], [async_add_one, async_add_one]))
        [2, 3]
    """
    input = to_list(input, flatten_dict, flat, dropna)
    func = to_list(func)
    assert len(input) == len(func), "Input and function counts must match."
    
    tasks = [al_call(inp, f, flatten_dict, flat, dropna) 
             for f, inp in zip(func, input)]
    out = await asyncio.gather(*tasks)
    return to_list(out, flat=True)

def e_call(input: Any, 
           func: Callable, 
           flatten_dict: bool = False, 
           flat: bool = False, 
           dropna: bool = True) -> List[Any]:
    """
    Applies each function in a list of functions to all elements in the input.

    This function expands the input to match the number of functions and then
    applies each function to the entire input. It is useful for applying a series
    of different transformations to the same input.

    Parameters:
        input (Union[Any, List[Any]]): The input or list of inputs to be processed.
        funcs (Union[Callable, List[Callable]]): The function or list of functions to apply.
        flatten_dict (bool, optional): Whether to flatten dictionaries in the input. Defaults to False.
        flat (bool, optional): Whether the output list should be flattened. Defaults to True.
        dropna (bool, optional): Whether to drop None values during flattening. Defaults to True.

    Returns:
        List[Any]: A list of results after applying each function to the input.

    Example:
        >>> def square(x):
        ...     return x**2
        >>> e_call([1, 2, 3], [square])
        [[1], [4], [9]]
    """

    f = lambda x, y: m_call(make_copy(x, len(to_list(y))), y, 
                            flatten_dict=flatten_dict, flat=flat, dropna=dropna)
    return to_list([f(inp, func) for inp in to_list(input)], flat=flat)

async def ae_call(input_: Any, 
                  func_: Callable, 
                  flatten_dict: bool = False,
                  flat: bool = False, 
                  dropna: bool = True):
    """
    Asynchronously applies each function in a list of functions to all elements in the input.

    This asynchronous function expands the input to match the number of functions and 
    then asynchronously applies each function to the entire input. It is useful for applying a series 
    of different asynchronous transformations to the same input.

    Parameters:
        input (Union[Any, List[Any]]): The input or list of inputs to be processed.
        funcs (Union[Callable, List[Callable]]): The function or list of functions to apply.
        flatten_dict (bool, optional): Whether to flatten dictionaries in the input. Defaults to False.
        flat (bool, optional): Whether the output list should be flattened. Defaults to True.
        dropna (bool, optional): Whether to drop None values during flattening. Defaults to True.

    Example:
        >>> async def async_square(x):
        ...     return x**2
        >>> asyncio.run(ae_call([1, 2, 3], [async_square]))
        [[1, 4, 9]]
    """
    async def async_f(x, y):
        return await am_call(make_copy(x, len(to_list(y))), y, flatten_dict=flatten_dict, flat=flat, dropna=dropna)

    tasks = [async_f(inp, func_) for inp in to_list(input_)]
    return await asyncio.gather(*tasks)

def get_timestamp() -> str:
    """
    Generates a current timestamp in a file-safe string format.

    This function creates a timestamp from the current time, formatted in ISO 8601 format, 
    and replaces characters that are typically problematic in filenames (like colons and periods) 
    with underscores.

    Returns:
        str: The current timestamp in a file-safe string format.

    Example:
        >>> get_timestamp()  # Doctest: +ELLIPSIS
        '...'
    """
    return datetime.now().isoformat().replace(":", "_").replace(".", "_")

def create_id() -> str:
    """
    Generates a unique ID based on the current time and random bytes.

    This function combines the current time in ISO 8601 format with 16 random bytes
    to create a unique identifier. The result is hashed using SHA-256 and the first
    16 characters of the hexadecimal digest are returned.

    Returns:
        str: A 16-character unique identifier.

    Example:
        >>> create_id()  # Doctest: +ELLIPSIS
        '...'
    """
    current_time = datetime.now().isoformat().encode('utf-8')
    random_bytes = os.urandom(16)
    return hashlib.sha256(current_time + random_bytes).hexdigest()[:16]

def create_path(dir: str, filename: str, timestamp: bool = True, dir_exist_ok: bool = True) -> str:
    """
    Creates a file path by optionally appending a timestamp to the filename.

    This function constructs a file path by combining a directory, an optional timestamp,
    and a filename. It also ensures the existence of the directory.

    Args:
        dir_ (str): The directory in which the file is to be located.
        filename (str): The name of the file.
        timestamp (bool, optional): If True, appends a timestamp to the filename. Defaults to True.
        dir_exist_ok (bool, optional): If True, creates the directory if it doesn't exist. Defaults to True.

    Returns:
        str: The full path to the file.

    Example:
        >>> create_path('/tmp/', 'log.txt', timestamp=False)
        '/tmp/log.txt'
    """
    os.makedirs(dir, exist_ok=dir_exist_ok)
    if timestamp:
        timestamp = get_timestamp()
        return f"{dir}{timestamp}{filename}"
    else:
        return f"{dir}{filename}"
    


import os
import unittest
import tempfile
from unittest.mock import patch
from pathlib import Path
from lionagi.utils.doc_util import (
    dir_to_path,
    read_text,
    dir_to_files,
    chunk_text,
    file_to_chunks,
    get_bins,
)


class Test_doc_util(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for testing
        self.temp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        # Clean up the temporary directory after each test
        self.temp_dir.cleanup()

    def test_dir_to_path(self):
        # Mock the 'glob' method to avoid actual file system operations
        with patch('lionagi.utils.doc_util.Path.glob') as mock_glob:
            mock_glob.return_value = [Path(self.temp_dir.name) / 'file1.txt', Path(self.temp_dir.name) / 'file2.txt']
            result = dir_to_path(dir=self.temp_dir.name, ext='.txt', recursive=True, flat=True)
        expected_result = [Path(self.temp_dir.name) / 'file1.txt', Path(self.temp_dir.name) / 'file2.txt']
        self.assertEqual(result, expected_result)

    def test_read_text(self):
        # Create a temporary file with known content
        with open(Path(self.temp_dir.name) / 'temp_test_file.txt', 'w') as temp_file:
            temp_file.write('Hello, World!')

        result = read_text(filepath=str(Path(self.temp_dir.name) / 'temp_test_file.txt'), clean=True)
        self.assertEqual(result, 'Hello, World!')

    def test_dir_to_files(self):
        with open(Path(self.temp_dir.name) / 'file1.txt', 'w') as file1:
            file1.write('Hello, World!')
        with open(Path(self.temp_dir.name) / 'file2.txt', 'w') as file2:
            file2.write('LionAGI is cool!')

        result = dir_to_files(dir=self.temp_dir.name, ext='.txt', to_csv=False)

        expected_result = [
            {'project': 'project', 'folder': os.path.basename(self.temp_dir.name), 'file': 'file2.txt', 'file_size': 16,
             'content': 'LionAGI is cool!'},
            {'project': 'project', 'folder': os.path.basename(self.temp_dir.name), 'file': 'file1.txt', 'file_size': 13,
             'content': 'Hello, World!'},
        ]
        self.assertEqual(result, expected_result)

    def test_chunk_text(self):
        result = chunk_text("This is a test string.", 10, 0.2, 1)
        expected_result =  ['This is a t', ' test string', 'ng.']
        self.assertEqual(result, expected_result)

    def test_file_to_chunks(self):
        # Mock chunk_text to avoid actual text chunking
        fileinput = [
            {'project': 'project', 'folder': os.path.basename(self.temp_dir.name), 'file': 'file2.txt', 'file_size': 16,
             'content': 'LionAGI is cool!'},
            {'project': 'project', 'folder': os.path.basename(self.temp_dir.name), 'file': 'file1.txt', 'file_size': 13,
             'content': 'Hello, World!'},
        ]
        result = file_to_chunks(fileinput, chunk_size=10, threshold=5)
        expected_result = [{'chunk_content': 'LionAGI is ', 'chunk_id': 1, 'chunk_overlap': 0.2, 'chunk_size': 11,
                            'chunk_threshold': 5, 'file': 'file2.txt', 'file_chunks': 2, 'file_size': 16,
                            'folder': os.path.basename(self.temp_dir.name), 'project': 'project'},
                           {'chunk_content': 's cool!', 'chunk_id': 2, 'chunk_overlap': 0.2, 'chunk_size': 7,
                            'chunk_threshold': 5, 'file': 'file2.txt', 'file_chunks': 2, 'file_size': 16,
                            'folder': os.path.basename(self.temp_dir.name), 'project': 'project'},
                           {'chunk_content': 'Hello, World!', 'chunk_id': 1, 'chunk_overlap': 0.2, 'chunk_size': 13,
                            'chunk_threshold': 5, 'file': 'file1.txt', 'file_chunks': 1, 'file_size': 13,
                            'folder': os.path.basename(self.temp_dir.name), 'project': 'project'}]
        self.assertEqual(result, expected_result)

    def test_get_bins(self):
        result = get_bins(['apple', 'a', 'b', 'banana', 'cheery', 'c', 'd', 'e'], upper=10)
        expected_result = [[0, 1, 2], [3], [4, 5, 6, 7]]
        self.assertEqual(result, expected_result)


if __name__ == '__main__':
    unittest.main()



from .sys_util import to_flat_dict, append_to_jsonl, to_list, str_to_num, make_copy, to_temp, to_csv, hold_call, ahold_call, l_call, al_call, m_call, am_call, e_call, ae_call, get_timestamp, _flatten_dict, _flatten_list, create_id, create_path
from .api_util import StatusTracker, RateLimiter, BaseAPIService, AsyncQueue
from .doc_util import dir_to_path, read_text, dir_to_files, chunk_text, file_to_chunks, file_to_chunks, get_bins
from .log_util import DataLogger

__all__ = [
    "to_list", "str_to_num", "make_copy", "to_temp", "to_csv", "hold_call", "ahold_call", "l_call", "al_call", "m_call", "am_call", "e_call", "ae_call", "get_timestamp", "_flatten_dict", "_flatten_list", "create_id", "create_path", "to_flat_dict", "append_to_jsonl",
    "StatusTracker", "RateLimiter", "BaseAPIService", "AsyncQueue",
    "dir_to_path", "read_text", "dir_to_files", "chunk_text", "file_to_chunks", "file_to_chunks", "get_bins",
    "DataLogger"
]


from collections import deque
from .sys_util import to_csv, create_path


class DataLogger:
    """
    Logs data entries and outputs them to a CSV file.
    
    This class provides a logging mechanism for data entries that can be saved to a
    CSV file. It offers methods for appending new log entries, saving the log to a CSV file,
    and setting the directory where the logs should be saved.

    Attributes:
        dir (str): The directory where the log files are to be saved.
        log (deque): A deque that stores log entries.

    Methods:
        __call__(entry): Appends a new entry to the log.
        to_csv(dir: str, filename: str, verbose: bool, timestamp: bool, dir_exist_ok: bool, file_exist_ok: bool): 
            Converts the log to a CSV format and saves it to a file.
        set_dir(dir: str): Sets the directory for saving log files.
    """

    def __init__(self, dir= None, log: list = None) -> None:
        """
        Initializes a new instance of the DataLogger class.

        Parameters:
            dir (str, optional): The directory where the log files will be saved. Defaults to None.
            log (list, optional): An initial list of log entries. Defaults to an empty deque.
        """
        self.dir = dir
        self.log = deque(log) if log else deque()

    def __call__(self, entry):
        """
        Appends a new entry to the log.

        Parameters:
            entry: The entry to append to the log. The entry can be of any datatype.
        """
        self.log.append(entry)

    def to_csv(self, dir: str, filename: str, verbose: bool = True, timestamp: bool = True, dir_exist_ok=True, file_exist_ok=False):
        """
        Converts the log to a CSV format and saves it to a file.

        Parameters:
            dir (str): The directory where the CSV file will be saved.
            filename (str): The name of the CSV file.
            verbose (bool, optional): If True, prints a message after saving the log. Defaults to True.
            timestamp (bool, optional): If True, appends a timestamp to the filename. Defaults to True.
            dir_exist_ok (bool, optional): If True, overrides the existing directory if needed. Defaults to True.
            file_exist_ok (bool, optional): If True, overrides the existing file if needed. Defaults to False.

        Postconditions:
            Saves the log entries to a CSV file and clears the `log` attribute.
            Optionally prints a message with the number of log entries saved and the file path.
        """
        filepath = create_path(dir=dir, filename=filename, timestamp=timestamp, dir_exist_ok=dir_exist_ok)
        to_csv(list(self.log), filepath, file_exist_ok=file_exist_ok)
        n_logs = len(list(self.log))
        self.log = deque()
        if verbose:
            print(f"{n_logs} logs saved to {filepath}")
            
    def set_dir(self, dir: str):
        """
        Sets the directory where log files will be saved.

        Parameters:
            dir (str): The directory to set for saving log files.
        """
        self.dir = dir
        


import asyncio
import logging
import tiktoken
from typing import Optional, NoReturn, Dict, Any

from ..utils.api_util import AsyncQueue, StatusTracker, RateLimiter, BaseAPIService


class OpenAIRateLimiter(RateLimiter):
    """
    A specialized RateLimiter for managing requests to the OpenAI API.

    Extends the generic RateLimiter to enforce specific rate-limiting rules and limits
    as required by the OpenAI API. This includes maximum requests and tokens per minute
    and replenishing these limits at regular intervals.

    Attributes:
        max_requests_per_minute (int): Maximum number of requests allowed per minute.
        max_tokens_per_minute (int): Maximum number of tokens allowed per minute.

    Methods:
        rate_limit_replenisher: Coroutine to replenish rate limits over time.
        calculate_num_token: Calculates the required tokens for a request.
    """

    def __init__(self, max_requests_per_minute: int, max_tokens_per_minute: int) -> None:
        """
        Initializes the rate limiter with specific limits for OpenAI API.

        Args:
            max_requests_per_minute (int): The maximum number of requests allowed per minute.
            max_tokens_per_minute (int): The maximum number of tokens that can accumulate per minute.
        """
        super().__init__(max_requests_per_minute, max_tokens_per_minute)
    
    async def rate_limit_replenisher(self) -> NoReturn:
        """
        Asynchronously replenishes the rate limit capacities at regular intervals.

        This coroutine runs indefinitely, replenishing the available request and token capacities
        based on the maximum limits defined for the OpenAI API. This task should run in the background
        to consistently reset the capacities.

        Example:
            >>> rate_limiter = OpenAIRateLimiter(100, 200)
            >>> asyncio.create_task(rate_limiter.rate_limit_replenisher())
            # This will start the background task for rate limit replenishment.
        """
        while True:
            await asyncio.sleep(60)  # Replenishes every 60 seconds
            self.available_request_capacity = self.max_requests_per_minute
            self.available_token_capacity = self.max_tokens_per_minute
            
    def calculate_num_token(self, payload: Dict[str, Any], 
                            api_endpoint: str, token_encoding_name: str) -> int:
        """
        Calculates the number of tokens required for a request based on the payload and API endpoint.

        The token calculation logic might vary based on different API endpoints and payload content.
        This method should be implemented in a subclass to provide the specific calculation logic
        for the OpenAI API.

        Args:
            payload (Dict[str, Any]): The payload of the request.
            api_endpoint (str): The specific API endpoint for the request.

        Returns:
            int: The estimated number of tokens required for the request.

        Example:
            >>> rate_limiter = OpenAIRateLimiter(100, 200)
            >>> payload = {'prompt': 'Translate the following text:', 'max_tokens': 50}
            >>> rate_limiter.calculate_num_token(payload, 'completions')
            # Expected token calculation for the given payload and endpoint.
        """

        encoding = tiktoken.get_encoding(token_encoding_name)
        if api_endpoint.endswith("completions"):
            max_tokens = payload.get("max_tokens", 15)
            n = payload.get("n", 1)
            completion_tokens = n * max_tokens

            # chat completions
            if api_endpoint.startswith("chat/"):
                num_tokens = 0
                for message in payload["messages"]:
                    num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
                    for key, value in message.items():
                        num_tokens += len(encoding.encode(value))
                        if key == "name":  # if there's a name, the role is omitted
                            num_tokens -= 1  # role is always required and always 1 token
                num_tokens += 2  # every reply is primed with <im_start>assistant
                return num_tokens + completion_tokens
            # normal completions
            else:
                prompt = payload["prompt"]
                if isinstance(prompt, str):  # single prompt
                    prompt_tokens = len(encoding.encode(prompt))
                    num_tokens = prompt_tokens + completion_tokens
                    return num_tokens
                elif isinstance(prompt, list):  # multiple prompts
                    prompt_tokens = sum([len(encoding.encode(p)) for p in prompt])
                    num_tokens = prompt_tokens + completion_tokens * len(prompt)
                    return num_tokens
                else:
                    raise TypeError(
                        'Expecting either string or list of strings for "prompt" field in completion request'
                    )
        elif api_endpoint == "embeddings":
            input = payload["input"]
            if isinstance(input, str):  # single input
                num_tokens = len(encoding.encode(input))
                return num_tokens
            elif isinstance(input, list):  # multiple inputs
                num_tokens = sum([len(encoding.encode(i)) for i in input])
                return num_tokens
            else:
                raise TypeError(
                    'Expecting either string or list of strings for "inputs" field in embedding request'
                )
        else:
            raise NotImplementedError(
                f'API endpoint "{api_endpoint}" not implemented in this script'
            )
            

class OpenAIService(BaseAPIService):
    """
    Service class for interacting with the OpenAI API.

    This class provides methods for calling OpenAI's API endpoints, handling the responses,
    and managing rate limits and asynchronous tasks associated with API calls.

    Attributes:
        base_url (str): The base URL for OpenAI's API.

    Methods:
        call_api: Call an API endpoint with a payload and handle the response.
    """

    base_url = "https://api.openai.com/v1/"

    def __init__(
        self,
        api_key: str,
        token_encoding_name: str,
        max_attempts: int,
        status_tracker: Optional[StatusTracker] = None,
        rate_limiter: Optional[OpenAIRateLimiter] = None,
        queue: Optional[AsyncQueue] = None
    ):
        """
        Initializes the OpenAI service with configuration for API interaction.

        Args:
            api_key (str): The API key for authenticating with OpenAI.
            token_encoding_name (str): The name of the text encoding used by OpenAI.
            max_attempts (int): The maximum number of attempts for calling an API endpoint.
            status_tracker (Optional[StatusTracker]): Tracker for API call outcomes.
            rate_limiter (Optional[OpenAIRateLimiter]): Rate limiter for OpenAI's limits.
            queue (Optional[AsyncQueue]): Queue for managing asynchronous API calls.

        Example:
            >>> service = OpenAIService(
            ...     api_key="api-key-123",
            ...     token_encoding_name="utf-8",
            ...     max_attempts=5,
            ...     status_tracker=None,
            ...     rate_limiter=OpenAIRateLimiter(100, 200),
            ...     queue=AsyncQueue()
            ... )
            # Service is configured for interacting with OpenAI API.
        """
        super().__init__(api_key, token_encoding_name, max_attempts, 
                         status_tracker, rate_limiter, queue)

    async def call_api(self, session, request_url: str, payload: Dict[str, any]) -> Optional[Dict[str, any]]:
        """
        Call an OpenAI API endpoint with a specific payload and handle the response.

        Args:
            session: The session object for making HTTP requests.
            request_url (str): The full URL of the OpenAI API endpoint to be called.
            payload (Dict[str, any]): The payload to send with the API request.

        Returns:
            Optional[Dict[str, any]]: The response data from the API call or None if the call fails.

        Raises:
            asyncio.TimeoutError: If the request attempts exceed the configured maximum limit.

        Example:
            >>> session = aiohttp.ClientSession()
            >>> response = await service.call_api(
            ...     session,
            ...     "https://api.openai.com/v1/engines",
            ...     {"model": "davinci"}
            ... )
            # Calls the specified API endpoint with the given payload.
        """
        while True:
            if self.rate_limiter.available_request_capacity < 1 or self.rate_limiter.available_token_capacity < 10:  # Minimum token count
                await asyncio.sleep(1)  # Wait for capacity
                continue
            
            endpoint = self.api_endpoint_from_url(request_url)
            required_tokens = self.rate_limiter.calculate_num_token(payload, endpoint, self.token_encoding_name)
            
            if self.rate_limiter.available_token_capacity >= required_tokens:
                self.rate_limiter.available_request_capacity -= 1
                self.rate_limiter.available_token_capacity -= required_tokens

                request_headers = {"Authorization": f"Bearer {self.api_key}"}
                attempts_left = self.max_attempts

                while attempts_left > 0:
                    try:
                        async with session.post(
                            url=request_url, headers=request_headers, json=payload
                        ) as response:
                            response_json = await response.json()

                            if "error" in response_json:
                                logging.warning(
                                    f"API call failed with error: {response_json['error']}"
                                )
                                attempts_left -= 1

                                if "Rate limit" in response_json["error"].get("message", ""):
                                    await asyncio.sleep(15)
                            else:
                                return response_json
                    except Exception as e:
                        logging.warning(f"API call failed with exception: {e}")
                        attempts_left -= 1

                logging.error("API call failed after all attempts.")
                break
            else:
                await asyncio.sleep(1)  # Wait for token capacity
                


import math
from pathlib import Path
from typing import Any, Dict, List, Union, Callable, Optional

from .sys_util import to_list, l_call
from .log_util import DataLogger


def dir_to_path(dir: str, ext, recursive: bool = False, flat: bool = True):
    """
    Retrieves a list of file paths in the specified directory with the given extension.

    Parameters:
        dir (str): The directory path where to search for files.
        ext (str): The file extension to filter by.
        recursive (bool, optional): If True, search for files recursively in subdirectories. Defaults to False.
        flat (bool, optional): If True, return a flat list of file paths. Defaults to True.

    Returns:
        List[str]: A list of file paths that match the given extension within the specified directory.

    Example:
        >>> files = dir_to_path(dir='my_directory', ext='.txt', recursive=True, flat=True)
    """

    def _dir_to_path(ext, recursive=recursive):
        tem = '**/*' if recursive else '*'
        return list(Path(dir).glob(tem + ext))
    
    return to_list(l_call(ext, _dir_to_path, flat=True), flat=flat)

def read_text(filepath: str, clean: bool = True) -> str:
    """
    Reads the content of a text file and optionally cleans it by removing specified characters.

    Parameters:
        filepath (str): The path to the text file to be read.
        clean (bool, optional): If True, clean the content by removing specific unwanted characters. Defaults to True.

    Returns:
        str: The cleaned (if 'clean' is True) or raw content of the text file.

    Example:
        >>> content = read_text(filepath='example.txt', clean=True)
    """

    with open(filepath, 'r') as f:
        content = f.read()
        if clean:
            # Define characters to replace and their replacements
            replacements = {'\\': ' ', '\\\n': ' ', '\\\t': ' ', '  ': ' ', '\'': ' '}
            for old, new in replacements.items():
                content = content.replace(old, new)
        return content

def dir_to_files(dir: str, ext: str, recursive: bool = False, 
                 reader: Callable = read_text, clean: bool = True, 
                 to_csv: bool = False, project: str = 'project',
                 output_dir: str = 'data/logs/sources/', filename: Optional[str] = None, 
                 verbose: bool = True, timestamp: bool = True, logger: Optional[DataLogger] = None):
    """
    Reads and processes files in a specified directory with the given extension.

    Parameters:
        dir (str): The directory path where files are located.
        ext (str): The file extension to filter by.
        recursive (bool, optional): If True, search files recursively in subdirectories. Defaults to False.
        reader (Callable, optional): Function used to read and process the content of each file. Defaults to read_text.
        clean (bool, optional): If True, cleans the content by removing specified characters. Defaults to True.
        to_csv (bool, optional): If True, export the processed data to a CSV file. Defaults to False.
        project (str, optional): The name of the project. Defaults to 'project'.
        output_dir (str, optional): Directory path for exporting the CSV file. Defaults to 'data/logs/sources/'.
        filename (Optional[str], optional): Name of the CSV file, if not provided, a default will be used. Defaults to None.
        verbose (bool, optional): If True, print a message upon CSV export. Defaults to True.
        timestamp (bool, optional): If True, include a timestamp in the file name. Defaults to True.
        logger (Optional[DataLogger], optional): An instance of DataLogger for logging, if not provided, a new one will be created. Defaults to None.

    Returns:
        List[Dict[str, Union[str, Path]]]: A list of dictionaries containing file information and content.

    Examples:
        >>> logs = dir_to_files(dir='my_directory', ext='.txt', to_csv=True)
    """
    
    sources = dir_to_path(dir, ext, recursive)
    
    def split_path(path: Path) -> tuple:
        folder_name = path.parent.name
        file_name = path.name
        return (folder_name, file_name)

    def to_dict(path_: Path) -> Dict[str, Union[str, Path]]:
        folder, file = split_path(path_)
        content = reader(str(path_), clean=clean)
        return {
            'project': project,
            'folder': folder,
            'file': file,
            "file_size": len(str(content)),
            'content': content
        } if content else None
    
    logs = to_list(l_call(sources, to_dict, flat=True), dropna=True)
    
    if to_csv:
        filename = filename or f"{project}_sources.csv"
        logger = DataLogger(dir=output_dir, log=logs) if not logger else logger
        logger.to_csv(dir=output_dir, filename=filename, verbose=verbose, timestamp=timestamp)

    return logs

def chunk_text(input: str, chunk_size: int, overlap: float, 
               threshold: int) -> List[Union[str, None]]:
    """
    Splits a string into chunks of a specified size, allowing for optional overlap between chunks.
    
    Args:
        input (str): The text to be split into chunks.
        chunk_size (int): The size of each chunk in characters.
        overlap (float): A value between [0, 1] specifying the percentage of overlap between adjacent chunks.
        threshold (int): The minimum size for the last chunk. If the last chunk is smaller than this, it will be merged with the previous chunk.
        
    Raises:
        TypeError: If input text cannot be converted to a string.
        ValueError: If any error occurs during the chunking process.
        
    Returns:
        List[Union[str, None]]: List of text chunks.
    """
    
    try:
        # Ensure text is a string
        if not isinstance(input, str):
            input = str(input)
        
        chunks = []
        n_chunks = math.ceil(len(input) / chunk_size)
        overlap_size = int(chunk_size * overlap / 2)
        
        if n_chunks == 1:
            return [input]
        
        elif n_chunks == 2:
            chunks.append(input[:chunk_size + overlap_size])
            if len(input) - chunk_size > threshold:
                chunks.append(input[chunk_size - overlap_size:])
            else:
                return [input]
            return chunks
        
        elif n_chunks > 2:
            chunks.append(input[:chunk_size + overlap_size])
            for i in range(1, n_chunks - 1):
                start_idx = chunk_size * i - overlap_size
                end_idx = chunk_size * (i + 1) + overlap_size
                chunks.append(input[start_idx:end_idx])
            
            if len(input) - chunk_size * (n_chunks - 1) > threshold:
                chunks.append(input[chunk_size * (n_chunks - 1) - overlap_size:])
            else:
                chunks[-1] += input[chunk_size * (n_chunks - 1):]
            
            return chunks
        
    except Exception as e:
        raise ValueError(f"An error occurred while chunking the text. {e}")

def _file_to_chunks(input: Dict[str, Any], 
                   field: str = 'content', 
                   chunk_size: int = 1500, 
                   overlap: float = 0.2, 
                   threshold: int = 200) -> List[Dict[str, Any]]:
    """
    Splits text from a specified dictionary field into chunks and returns a list of dictionaries.
    
    Args:
        input (Dict[str, Any]): The input dictionary containing the text field to be chunked.
        field (str, optional): The dictionary key corresponding to the text field. Defaults to 'content'.
        chunk_size (int, optional): Size of each text chunk in characters. Defaults to 1500.
        overlap (float, optional): Percentage of overlap between adjacent chunks, in the range [0, 1]. Defaults to 0.2.
        threshold (int, optional): Minimum size for the last chunk. If smaller, it will be merged with the previous chunk. Defaults to 200.
        
    Raises:
        ValueError: If any error occurs during the chunking process.
        
    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each containing a separate chunk along with original key-value pairs from the input dictionary.
    
    Example:
        >>> d = {'content': 'This is a test string.', 'other_field': 1}
        >>> file_to_chunks(d)
        [{'chunk_overlap': 0.2, 'chunk_threshold': 200, 'file_chunks': 2, 'chunk_id': 1, 'chunk_size': 14, 'chunk_content': 'This is a test', 'other_field': 1}, ...]
    """

    try:
        out = {key: value for key, value in input.items() if key != field}
        out.update({"chunk_overlap": overlap, "chunk_threshold": threshold})
        
        chunks = chunk_text(input[field], chunk_size=chunk_size, overlap=overlap, threshold=threshold)
        logs = []
        for i, chunk in enumerate(chunks):
            chunk_dict = out.copy()
            chunk_dict.update({
                'file_chunks': len(chunks),
                'chunk_id': i + 1,
                'chunk_size': len(chunk),
                f'chunk_{field}': chunk
            })
            logs.append(chunk_dict)

        return logs
        
    except Exception as e:
        raise ValueError(f"An error occurred while chunking the file. {e}")
    
def file_to_chunks(input, 
                   field: str = 'content', 
                   chunk_size: int = 1500, 
                   overlap: float = 0.2, 
                   threshold: int = 200, 
                   to_csv=False, 
                   project='project',
                   output_dir='data/logs/sources/', 
                   chunk_func = _file_to_chunks,
                   filename=None, 
                   verbose=True, 
                   timestamp=True, 
                   logger=None):
    """
        Splits text from a specified dictionary field into chunks and returns a list of dictionaries.

    Args:
        input (List[Dict[str, Any]]): The input dictionaries containing the text field to be chunked.
        field (str, optional): The dictionary key corresponding to the text field. Defaults to 'content'.
        chunk_size (int, optional): Size of each text chunk in characters. Defaults to 1500.
        overlap (float, optional): Percentage of overlap between adjacent chunks, in the range [0, 1]. Defaults to 0.2.
        threshold (int, optional): Minimum size for the last chunk. If smaller, it will be merged with the previous chunk. Defaults to 200.
        to_csv: If True, export the processed data to a CSV file.
        project: The name of the project.
        output_dir: The directory path for exporting the CSV file.
        filename: The name of the CSV file.
        verbose: If True, print a verbose message after export.
        timestamp: If True, include a timestamp in the exported file name.
        logger: An optional DataLogger instance for logging.
    """
    
    f = lambda x: chunk_func(x, field=field, chunk_size=chunk_size, overlap=overlap, threshold=threshold)
    logs = to_list(l_call(input, f), flat=True)
    
    if to_csv:
        filename = filename if filename else f"{project}_sources.csv"
        logger = DataLogger(log=logs) if not logger else logger
        logger.to_csv(dir=output_dir, filename=filename, verbose=verbose, timestamp=timestamp)

    return logs

def get_bins(input: List[str], upper: int = 7500) -> List[List[int]]:
    """
    Get index of elements in a list based on their consecutive cumulative sum of length,
    according to some upper threshold. Return lists of indices as bins.
    
    Args:
    input (List[str]): List of items to be binned.
    upper (int, optional): Upper threshold for the cumulative sum of the length of items in a bin. Default is 7500.
    
    Returns:
    List[List[int]]: List of lists, where each inner list contains the indices of the items that form a bin.
    
    Example:
    >>> items = ['apple', 'a', 'b', 'banana', 'cheery', 'c', 'd', 'e']
    >>> upper = 10
    >>> get_bins(items, upper)
    [[0, 1, 2], [3], [4, 5, 6, 7]]
    """
    current = 0
    bins = []
    bin = []
    for idx, item in enumerate(input):
        if current + len(item) < upper:
            bin.append(idx)
            current += len(item)
        elif current + len(item) >= upper:
            bins.append(bin)
            bin = [idx]
            current = len(item)
        if idx == len(input) - 1 and len(bin) > 0:
            bins.append(bin)
    return bins



from .message import Message

class Conversation:
    """
    A class modeling conversations and managing messages within the conversation.

    This class facilitates the organization and manipulation of messages in a conversation.
    It includes methods for initiating a conversation, adding messages, changing the system message,
    appending the last response, and keeping the last N exchanges.

    Attributes:
        response_counts: A class-level attribute to track the number of responses.
        messages: A list containing messages in the conversation.
        msg: An instance of the Message class for creating and processing messages.
        responses: A list containing response messages.

    Methods:
        initiate_conversation: Start a new conversation with system and user instructions.
        add_messages: Add messages to the conversation.
        change_system: Change the system message in the conversation.
        append_last_response: Append the last response to the conversation.
        keep_last_n_exchanges: Keep only the last N exchanges in the conversation.
    """
    
    response_counts = 0
    
    def __init__(self, messages=None) -> None:
        """
        Initialize a Conversation object.

        Args:
            messages: A list of messages to initialize the conversation.
        """
        self.messages = messages or []
        self.msg = Message()
        self.responses = []

    def initiate_conversation(self, system, instruction, context=None):
        """
        Start a new conversation with a system message and an instruction.

        Args:
            system: The content of the system message.
            instruction: The content of the user instruction.
            context: Additional context for the user instruction.
        """
        self.messages, self.responses = [], []
        self.add_messages(system=system)
        self.add_messages(instruction=instruction, context=context)

    def add_messages(self, system=None, instruction=None, context=None, response=None):
        """
        Add messages to the conversation.

        Args:
            system: The content of the system message.
            instruction: The content of the user instruction.
            context: Additional context for the user instruction.
            response: The content of the assistant's response.
        """
        msg = self.msg(system=system, instruction=instruction, response=response, context=context)
        self.messages.append(msg)

    def change_system(self, system):
        """
        Change the system message in the conversation.

        Args:
            system: The new content for the system message.
        """
        self.messages[0] = self.msg(system=system)

    def append_last_response(self):
        """
        Append the last response to the conversation.
        """
        self.add_messages(response=self.responses[-1])

    def keep_last_n_exchanges(self, n: int):
        """
        Keep only the last N exchanges in the conversation.

        Args:
            n: The number of exchanges to keep.
        """
        response_indices = [
            index for index, message in enumerate(self.messages[1:]) if message["role"] == "assistant"
        ]
        if len(response_indices) >= n:
            first_index_to_keep = response_indices[-n] + 1
            self.messages = [self.system] + self.messages[first_index_to_keep:]


import asyncio
import json
import logging
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, NoReturn, Optional

@dataclass
class StatusTracker:
    """Class for keeping track of various task statuses.
    
    This class serves as a simple way to monitor different types of task
    outcomes and errors within a system. It uses dataclasses for easy
    creation and management of state.
    
    Attributes:
        num_tasks_started: The number of tasks that have been initiated.
        num_tasks_in_progress: The number of tasks currently being processed.
        num_tasks_succeeded: The number of tasks that have completed successfully.
        num_tasks_failed: The number of tasks that have failed.
        num_rate_limit_errors: The number of tasks that failed due to rate limiting.
        num_api_errors: The number of tasks that failed due to API errors.
        num_other_errors: The number of tasks that failed due to other errors.
    """
    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0
    num_other_errors: int = 0


class AsyncQueue:
    """
    A queue class that handles asynchronous operations using asyncio.

    This class provides an asynchronous queue that can enqueue items, process them
    asynchronously, and support graceful shutdowns. It is designed to facilitate
    concurrent task processing in an orderly and controlled manner.

    Attributes:
        queue (asyncio.Queue): A queue to hold items for asynchronous processing.
        _stop_event (asyncio.Event): An event to signal when the queue should stop processing.

    Methods:
        enqueue(item): Add an item to the queue for processing.
        dequeue(): Remove and return an item from the queue.
        join(): Wait until all items in the queue have been processed.
        stop(): Signal to stop processing new items in the queue.
        stopped(): Check if the queue has been signaled to stop.
        process_requests(func): Process items using a provided function.
    """

    def __init__(self) -> None:
        """
        Initializes an AsyncQueue object with an empty asyncio Queue and a stop event.
        """
        self.queue = asyncio.Queue()
        self._stop_event = asyncio.Event()

    async def enqueue(self, item: Any) -> None:
        """
        Asynchronously add an item to the queue for processing.

        Args:
            item (Any): The item to be added to the queue.

        Example:
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.enqueue('Task 1'))
        """
        await self.queue.put(item)

    async def dequeue(self) -> Any:
        """
        Asynchronously remove and return an item from the queue.

        If the queue is empty, this method will wait until an item is available.

        Returns:
            Any: The next item from the queue.

        Example:
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.enqueue('Task 1'))
            >>> asyncio.run(async_queue.dequeue())
            'Task 1'
        """
        return await self.queue.get()

    async def join(self) -> None:
        """
        Asynchronously wait until all items in the queue have been processed.

        This method blocks until every item that has been enqueued is processed, 
        ensuring that all tasks are completed.

        Example:
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.enqueue('Task 1'))
            >>> asyncio.run(async_queue.join())  # This will block until 'Task 1' is processed.
        """
        await self.queue.join()

    async def stop(self) -> None:
        """
        Signal the queue to stop processing new items.

        Once called, the queue will not process any new items after the current ones
        are completed, allowing for a graceful shutdown.

        Example:
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.stop())  # This signals the queue to stop processing.
        """
        self._stop_event.set()

    def stopped(self) -> bool:
        """
        Check if the queue has been signaled to stop processing.

        Returns:
            bool: True if a stop has been signaled, False otherwise.

        Example:
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.stop())
            >>> async_queue.stopped()
            True
        """
        return self._stop_event.is_set()

    async def process_requests(self, func: Callable[[Any], Any]) -> None:
        """
        Asynchronously process items from the queue using the provided function.

        Continuously dequeues items and applies the given function to each. 
        The processing stops when the queue is signaled to stop or a sentinel value (`None`) is dequeued.

        Args:
            func (Callable[[Any], Any]): A coroutine function to process items from the queue.

        Example:
            >>> async def sample_processing(task):
            ...     print("Processing:", task)
            >>> async_queue = AsyncQueue()
            >>> asyncio.run(async_queue.enqueue('Task 1'))
            >>> asyncio.run(async_queue.process_requests(sample_processing))
            Processing: Task 1
        """
        while not self.stopped():
            item = await self.dequeue()
            if item is None:  # Using `None` as a sentinel value to cease processing.
                await self.stop()
                break
            await func(item)


class RateLimiter(ABC):
    """
    An abstract base class for rate limiting mechanisms.

    This class defines a structure for rate limiters, which are used to control the frequency
    of requests sent to or received from a network interface controller or an API.

    Attributes:
        max_requests_per_minute (int): Maximum number of requests permitted per minute.
        max_tokens_per_minute (int): Maximum number of tokens that can accumulate per minute.
        available_request_capacity (int): Current number of available request slots.
        available_token_capacity (int): Current number of available tokens.
        rate_limit_replenisher_task (asyncio.Task): Background task for replenishing rate limits.

    Methods:
        rate_limit_replenisher: Coroutine to replenish rate limits over time.
        calculate_num_token: Method to calculate required tokens for a request.
    """
    
    def __init__(self, max_requests_per_minute: int, max_tokens_per_minute: int) -> None:
        """
        Initializes the RateLimiter with specified maximum request and token limits.

        Args:
            max_requests_per_minute (int): Maximum requests allowed per minute.
            max_tokens_per_minute (int): Maximum tokens allowed to accumulate per minute.

        Example:
            >>> class MyRateLimiter(RateLimiter):
            ...     async def rate_limit_replenisher(self) -> NoReturn:
            ...         # Implementation for rate replenishment.
            ...     def calculate_num_token(self, payload: Dict[str, Any], api_endpoint: str) -> int:
            ...         # Implementation for token calculation.
            ...
            >>> limiter = MyRateLimiter(100, 200)
        """
        self.max_requests_per_minute = max_requests_per_minute
        self.max_tokens_per_minute = max_tokens_per_minute
        self.available_request_capacity = max_requests_per_minute
        self.available_token_capacity = max_tokens_per_minute
        self.rate_limit_replenisher_task = asyncio.create_task(self.rate_limit_replenisher())
    
    @abstractmethod
    async def rate_limit_replenisher(self) -> NoReturn:
        """
        Asynchronously replenishes rate limit capacities.

        This coroutine should be implemented to periodically restore `available_request_capacity`
        and `available_token_capacity` according to specific rules defined in subclasses.

        Example:
            >>> class MyRateLimiter(RateLimiter):
            ...     async def rate_limit_replenisher(self) -> NoReturn:
            ...         while True:
            ...             # Replenishment logic here
            ...
            >>> limiter = MyRateLimiter(100, 200)
        """
        pass
    
    @abstractmethod
    def calculate_num_token(self, payload: Dict[str, Any], api_endpoint: str) -> int:
        """
        Calculates required tokens for a request.

        Subclasses should implement this method to determine the number of tokens needed based
        on the request payload and target endpoint.

        Args:
            payload (Dict[str, Any]): Payload of the request.
            api_endpoint (str): Target API endpoint for the request.

        Returns:
            int: Calculated number of tokens required for the request.

        Example:
            >>> class MyRateLimiter(RateLimiter):
            ...     def calculate_num_token(self, payload: Dict[str, Any], api_endpoint: str) -> int:
            ...         return len(payload.get('data', '')) // 10
            ...
            >>> limiter = MyRateLimiter(100, 200)
            >>> limiter.calculate_num_token({'data': '12345'}, 'api/send')
            0
        """
        pass

class BaseAPIService(ABC):
    """
    Abstract base class for API services requiring asynchronous operations.

    This class provides foundational attributes and methods for error handling,
    logging, rate-limiting, and re-attempting API calls. It is designed to be
    subclassed for concrete implementations of specific API service interactions.

    Attributes:
        api_key (str): The API key used for authenticating with the API service.
        token_encoding_name (str): The encoding for the API token.
        max_attempts (int): The maximum number of retry attempts for API calls.
        status_tracker (StatusTracker): Tracker for API call statuses.
        rate_limiter (RateLimiter): Limiter to control the rate of API calls.
        queue (AsyncQueue): Queue for managing API call tasks.

    Methods:
        call_api: Abstract method to define API call mechanism in subclasses.
        handle_error: Handle errors by logging and saving details to a JSONL file.
        append_to_jsonl: Append data to a file in JSONL format.
        api_endpoint_from_url: Extract the API endpoint from a URL.
        task_id_generator_function: Generate a sequence of unique task IDs.
    """
    
    def __init__(
        self,
        api_key: str,
        token_encoding_name: str,
        max_attempts: int,
        status_tracker: Optional[StatusTracker],
        rate_limiter: RateLimiter,
        queue: Optional[AsyncQueue]
    ) -> None:
        """
        Initializes the BaseAPIService with necessary configuration.

        Args:
            api_key (str): The API key for authentication.
            token_encoding_name (str): Encoding name for the API token.
            max_attempts (int): Maximum number of attempts for an API call.
            status_tracker (Optional[StatusTracker]): Tracker for API call statuses.
            rate_limiter (RateLimiter): Limiter for API call rates.
            queue (Optional[AsyncQueue]): Queue for managing API tasks.

        Example:
            >>> class MyAPIService(BaseAPIService):
            ...     # Implementation details here
            ...
            >>> service = MyAPIService(api_key="12345", token_encoding_name="utf-8",
            ...                       max_attempts=3, status_tracker=None,
            ...                       rate_limiter=my_rate_limiter, queue=None)
        """
        self.api_key = api_key
        self.token_encoding_name = token_encoding_name
        self.max_attempts = max_attempts
        self.status_tracker = status_tracker or StatusTracker()
        self.rate_limiter = rate_limiter
        self.queue = queue or AsyncQueue()
        
    @abstractmethod
    async def call_api(self) -> Any:
        """
        Abstract method to be implemented for making specific API calls.

        This method should define the logic for interacting with an API endpoint
        and must be implemented in subclasses.

        Example:
            >>> class MyAPIService(BaseAPIService):
            ...     async def call_api(self):
            ...         # Implementation details here
            ...
        """
        pass

    def handle_error(
        self,
        error: Exception,
        payload: Any,
        metadata: Any,
        save_filepath: str
    ) -> None:
        """
        Handles exceptions that occur during the API call process.
        
        Updates the status tracker to indicate the error and saves details to a JSONL file.
        
        Args:
            error (Exception): The exception that was raised during the API call.
            payload (Any): The data payload that was used for the API call.
            metadata (Any): Additional metadata related to the API call.
            save_filepath (str): The file path where error details should be saved.
        """
        self.status_tracker.num_tasks_in_progress -= 1
        self.status_tracker.num_tasks_failed += 1
        data = (
            [payload, [str(error)], metadata] 
            if metadata 
            else [payload, [str(error)]]
        )
        self.append_to_jsonl(data, save_filepath)
        logging.error(f"Request failed after all attempts. Saving errors: {data}")

    @staticmethod
    def append_to_jsonl(data: Any, filename: str) -> None:
        """
        Appends the given data to the specified JSONL file.
        
        Args:
            data (Any): The data to be appended in JSON Lines format.
            filename (str): The file path to the JSONL file.
        """
        json_string = json.dumps(data)
        with open(filename, "a") as f:
            f.write(json_string + "\n")

    @staticmethod
    def api_endpoint_from_url(request_url: str) -> str:
        """
        Extracts the endpoint from an API request URL.
        
        Args:
            request_url (str): The URL from which to extract the API endpoint.
        
        Returns:
            str: The extracted API endpoint.

        Example:
            >>> BaseAPIService.api_endpoint_from_url("https://api.example.com/v1/test_endpoint")
            'test_endpoint'
        """
        match = re.search(r"^https://[^/]+/v\d+/(.+)$", request_url)
        if match:
            return match.group(1)
        else:
            return ""

    @staticmethod
    def task_id_generator_function() -> Generator[int, None, None]:
        """
        Generates a continuous sequence of integers for task IDs.
        
        Yields:
            int: The next task ID in the sequence (0, 1, 2, ...).

        Example:
            >>> id_gen = BaseAPIService.task_id_generator_function()
            >>> next(id_gen)
            0
            >>> next(id_gen)
            1
        """
        task_id = 0
        while True:
            yield task_id
            task_id += 1



import unittest
import asyncio
import os
from tempfile import NamedTemporaryFile
from lionagi.utils.sys_util import (
    to_flat_dict,
    to_list,
    str_to_num,
    make_copy,
    to_temp,
    to_csv,
    append_to_jsonl,
    hold_call,
    ahold_call,
    l_call,
    al_call,
    m_call,
    am_call,
    e_call,
    ae_call,
    get_timestamp,
    create_id,
    create_path,
)


class Test_sys_util(unittest.TestCase):
    def test_to_flat_dict(self):
        example_nested_dict = {'a': 1, 'b': {'c': 2}}
        result = to_flat_dict(example_nested_dict)
        self.assertEqual(result, {'a': 1, 'b_c': 2})

    def test_to_list(self):
        example_dict = {'a': 1, 'b': [2, 3]}
        result = to_list(example_dict, flatten_dict=True)
        self.assertEqual(result, [{'a': 1}, {'b_0': 2}, {'b_1': 3}])

    def test_str_to_num(self):
        result = str_to_num("Temperature is -5.6 degrees", num_type=float, precision=1)
        self.assertEqual(result, -5.6)

    def test_make_copy(self):
        sample_dict = {'key': 'value'}
        result = make_copy(sample_dict, 2)
        self.assertEqual(result, [{'key': 'value'}, {'key': 'value'}])

    # def test_to_temp(self):
    #     temp_file = to_temp({'a': 1, 'b': [2, 3]}, flatten_dict=True)
    #     self.assertTrue(isinstance(temp_file, NamedTemporaryFile))

    def test_to_csv(self):
        data = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
        filepath = 'test_people.csv'
        to_csv(data, filepath, file_exist_ok=True)
        self.assertTrue(os.path.exists(filepath))

    # def test_append_to_jsonl(self):
    #     data = {"key": "value"}
    #     filepath = 'test_data.jsonl'
    #     append_to_jsonl(data, filepath)
    #     with open(filepath, 'r') as f:
    #         lines = f.readlines()
    #         self.assertTrue(len(lines) == 1)
    #         self.assertEqual(lines[0].strip(), '{"key": "value"}')

    def test_hold_call(self):
        def add_one(x):
            return x + 1

        result = hold_call(5, add_one, sleep=2)
        self.assertEqual(result, 6)

    def test_ahold_call(self):
        async def async_add_one(x):
            return x + 1

        result = asyncio.run(ahold_call(5, async_add_one, sleep=2))
        self.assertEqual(result, 6)

    def test_l_call(self):
        def square(x):
            return x * x

        result = l_call([1, 2, 3], square)
        self.assertEqual(result, [1, 4, 9])

    def test_al_call(self):
        async def async_square(x):
            return x * x

        result = asyncio.run(al_call([1, 2, 3], async_square))
        self.assertEqual(result, [1, 4, 9])

    def test_m_call(self):
        def add_one(x):
            return x + 1

        result = m_call([1, 2], [add_one, add_one])
        self.assertEqual(result, [2, 3])

    def test_am_call(self):
        async def async_add_one(x):
            return x + 1

        result = asyncio.run(am_call([1, 2], [async_add_one, async_add_one]))
        self.assertEqual(result, [2, 3])

    def test_e_call(self):
        def square(x):
            return x**2

        result = e_call([1, 2, 3], [square])
        self.assertEqual(result, [[1], [4], [9]])

    def test_ae_call(self):
        async def async_square(x):
            return x**2

        result = asyncio.run(ae_call([1, 2, 3], [async_square]))
        self.assertEqual(result, [[1], [4], [9]])

    def test_get_timestamp(self):
        result = get_timestamp()
        self.assertTrue(isinstance(result, str))

    def test_create_id(self):
        result = create_id()
        self.assertTrue(isinstance(result, str))
        self.assertEqual(len(result), 16)

    def test_create_path(self):
        result = create_path('/tmp/', 'log.txt', timestamp=False)
        self.assertEqual(result, '/tmp/log.txt')


if __name__ == '__main__':
    unittest.main()



import os
import dotenv

dotenv.load_dotenv()

from ..api.OAIService import OpenAIService, OpenAIRateLimiter

# default to openai tier 1, gpt4-1106-preview
OAIRateLimiter = OpenAIRateLimiter(
    max_requests_per_minute = 500,
    max_tokens_per_minute = 150000
)

OAIService = OpenAIService(
    api_key = os.getenv("OPENAI_API_KEY"),
    token_encoding_name = "cl100k_base",
    max_attempts=5,
    rate_limiter=OAIRateLimiter,
)


import unittest
from lionagi.utils.log_util import DataLogger
from unittest.mock import patch
import os
import csv
import tempfile

class TestDataLogger(unittest.TestCase):
    """
    Unit tests for the DataLogger class.
    """

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Clean up the temporary directory after each test
        os.rmdir(self.temp_dir)

    def test_init(self):
        logger = DataLogger(dir='test_dir', log=[1, 2, 3])
        self.assertEqual(logger.dir, 'test_dir')
        self.assertListEqual(list(logger.log), [1, 2, 3])

    def test_call(self):
        logger = DataLogger()
        logger('new entry')
        self.assertListEqual(list(logger.log), ['new entry'])

    def test_to_csv(self):
        logger = DataLogger(dir=self.temp_dir, log=[{'id': 1, 'value': 10}, {'id': 2, 'value':20}])

        with patch('builtins.print') as mock_print:
            logger.to_csv(dir=self.temp_dir, filename='test.csv', timestamp=False)

        csv_filepath = self.temp_dir+'test.csv'
        mock_print.assert_called_with(f"2 logs saved to {csv_filepath}")

        self.assertTrue(os.path.exists(csv_filepath))

        with open(csv_filepath, 'r') as csv_file:
            csv_reader = csv.reader(csv_file)
            rows = list(csv_reader)
            self.assertEqual(len(rows), 3)
            self.assertListEqual(rows[0], ['id', 'value'])
            self.assertListEqual(rows[1], ['1', '10'])
            self.assertListEqual(rows[2], ['2', '20'])

        self.assertListEqual(list(logger.log), [])

    def test_set_dir(self):
        logger = DataLogger()
        logger.set_dir('new_test_dir')
        self.assertEqual(logger.dir, 'new_test_dir')

if __name__ == '__main__':
    unittest.main()



llmconfig = {
    "model": "gpt-4-1106-preview",
    "frequency_penalty": 0,
    "max_tokens": 4000,
    "n": 1,
    "presence_penalty": 0, 
    "response_format": {"type": "text"}, 
    "seed": None, 
    "stop": None,
    "stream": False,
    "temperature": 0.7, 
    "top_p": 1, 
    "tools": None,
    "tool_choice": "none", 
    "user": None
    }


from datetime import datetime
import json
from ..utils.sys_util import create_id
from ..utils.log_util import DataLogger


class Message:
    """
    A class modeling messages in conversations.

    This class is designed to encapsulate messages exchanged between different roles
    (user, assistant, or system) in a conversation. It includes functionality to process,
    log, and manage messages.

    Attributes:
        role: The role of the message (user, assistant, or system).
        content: The content of the message.
        sender: The sender of the message.
        logger: An instance of DataLogger for logging message information.

    Methods:
        __call__: Process and log a message based on its role.
    """
    def __init__(self) -> None:
        """
        Initialize a Message object with attributes for role, content, sender, and a DataLogger.
        """
        self.role = None
        self.content = None
        self.sender = None
        self.logger = DataLogger()

    def __call__(self, system=None, instruction=None, response=None, context=None, sender=None):
        """
        Process and log a message based on the specified role (system, instruction, or response).

        Args:
            system: The content of the message in the system role.
            instruction: The content of the message in the user role.
            response: The content of the message in the assistant role.
            context: Additional context for the user instruction.
            sender: The sender of the message.
        """
        
        if sum(map(bool, [system, instruction, response])) > 1:
            raise ValueError("Message cannot have more than one role.")
        else:
            if response:
                self.role = "assistant"
                self.sender = sender or "assistant"
                self.content = response['content']
            elif instruction:
                self.role = "user"
                self.sender = sender or "user"
                self.content = {"instruction": instruction}
                if context:
                    self.content.update(context)
            elif system:
                self.role = "system"
                self.sender = sender or "system"
                self.content = system
        out = {
            "role": self.role,
            "content": json.dumps(self.content) if isinstance(self.content, dict) else self.content
            }
        
        a = {**{
            "id": create_id(),
            "timestamp": datetime.now().isoformat(),
            "sender": self.sender
        }, **out}
        self.logger(a)
        return out


import aiohttp
import asyncio
from typing import Any

from .conversation import Conversation
from ..config.llmconfig import llmconfig
from ..utils.log_util import DataLogger
from ..utils.api_util import StatusTracker
from ..config.oaiconfig import OAIService

status_tracker = StatusTracker()

class Session():
    """
    A class representing a conversation session with chat completion capabilities.

    This class manages conversations, interacts with chat completion services (currently OpenAI),
    and logs the interactions using a DataLogger.

    Attributes:
        conversation: An instance of the Conversation class for managing messages.
        system: The system identifier for the conversation session.
        llmconfig: Configuration parameters for language models.
        logger: An instance of DataLogger for logging conversation interactions.
        api_service: An instance of the API service for making asynchronous API calls.

    Methods:
        initiate: Initiate a conversation session with the given instruction.
        followup: Continue the conversation session with a follow-up instruction.
        create_payload_chatcompletion: Create a payload for chat completion API calls.
        call_chatcompletion: Make an asynchronous call to the chat completion API.
    """
    
    def __init__(self, system, dir=None, llmconfig=llmconfig, api_service=OAIService):
        """
        Initialize a Session object.

        Args:
            system: The system identifier for the conversation session.
            dir: The directory for logging interactions.
            llmconfig: Configuration parameters for language models.
            api_service: An instance of the API service for making asynchronous API calls.
        """
        self.conversation = Conversation()
        self.system = system
        self.llmconfig = llmconfig
        self.logger = DataLogger(dir=dir)
        self.api_service = api_service
    
    async def initiate(self, instruction, system=None, context=None, out=True, **kwargs) -> Any:
        """
        Initiate a conversation session with the given instruction.

        Args:
            instruction: The user's instruction to initiate the conversation.
            system: The content of the system message.
            context: Additional context for the user instruction.
            out: Whether to return the output content.

        Returns:
            Any: The output content if 'out' is True, otherwise None.
        """
        config = {**self.llmconfig, **kwargs}
        system = system or self.system
        self.conversation.initiate_conversation(system=system, instruction=instruction, context=context)
        
        await self.call_chatcompletion(**config)
        if out:
            return self.conversation.responses[-1]['content']        

    async def followup(self, instruction, system=None, context=None, out=True, **kwargs) -> Any:
        """
        Continue the conversation session with a follow-up instruction.

        Args:
            instruction: The user's follow-up instruction.
            system: The content of the system message.
            context: Additional context for the user instruction.
            out: Whether to return the output content.

        Returns:
            Any: The output content if 'out' is True, otherwise None.
        """
        self.conversation.append_last_response()
        if system:
            self.conversation.change_system(system)
        self.conversation.add_messages(instruction=instruction, context=context)
        
        config = {**self.llmconfig, **kwargs}
        await self.call_chatcompletion(**config)
        if out:
            return self.conversation.responses[-1]['content']
    
    def create_payload_chatcompletion(self, **kwargs):
        """
        Create a payload for chat completion API calls.

        Args:
            kwargs: Additional keyword arguments for customization.
        """
        # currently only openai chat completions are supported
        messages = self.conversation.messages
        request_url = f"https://api.openai.com/v1/chat/completions"
        config = {**self.llmconfig, **kwargs}
        
        payload = {
            "messages": messages,
            "model": config.get('model'),
            "frequency_penalty": config.get('frequency_penalty'),
            "max_tokens": config.get('max_tokens'),
            "n": config.get('n'),
            "presence_penalty": config.get('presence_penalty'),
            "response_format": config.get('response_format'),
            "temperature": config.get('temperature'),
            "top_p": config.get('top_p'),
            }
        
        for key in ["seed", "stop", "stream", "tools", "tool_choice", "user"]:
            if config[key] is True:
                payload.update({key: config[key]})
    
        return (payload, request_url)
    
    async def call_chatcompletion(self, delay=1, **kwargs):
        """
        Make an asynchronous call to the chat completion API.

        Args:
            delay: The delay (in seconds) between API calls.
            kwargs: Additional keyword arguments for customization.
        """
        # currently only openai chat completions are supported
        payload, request_url = self.create_payload_chatcompletion(**kwargs)
        try:
            async with aiohttp.ClientSession() as session:
                completion = await self.api_service.call_api(session, request_url, payload)
                if "choices" in completion:
                    completion = completion['choices'][0]       # currently can only call one completion at a time, n has to be 1
                    self.logger({"input":self.conversation.messages, "output": completion})
                    response = {"role": "assistant", "content": completion['message']["content"]}
                    self.conversation.responses.append(response)
                    self.conversation.response_counts += 1
                    await asyncio.sleep(delay=delay)
                    status_tracker.num_tasks_succeeded += 1
                else:
                    status_tracker.num_tasks_failed += 1
        except Exception as e:
            status_tracker.num_tasks_failed += 1
            raise e